#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-


################################################
#############   MACHINE LEARNING   #############
##############     -  2016 -    ################
################################################

"""
Contact: Adrian Ulges (adrian.ulges@hs-rm.de)
"""

# ==============================================================================
# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import argparse
import cv2
import numpy as np
import time

# Import data
from tensorflow.examples.tutorials.mnist import input_data
import tensorflow as tf

FLAGS = None
HEIGHT = 28
WIDTH = 28
NDIMS = HEIGHT*WIDTH
NCLASS = 10

def model_softmax():

  # Create the model
  x = tf.placeholder(tf.float32, [None, NDIMS])
  W = tf.Variable(tf.zeros([NDIMS, NCLASS]))
  b = tf.Variable(tf.zeros([NCLASS]))
  y = tf.matmul(x, W) + b

  # Define loss and optimizer
  y_ = tf.placeholder(tf.float32, [None, 10])

  # The raw formulation of cross-entropy,
  #
  #   tf.reduce_mean(-tf.reduce_sum(y_ * tf.log(tf.softmax(y)),
  #                                 reduction_indices=[1]))
  #
  # can be numerically unstable.
  #
  # So here we use tf.nn.softmax_cross_entropy_with_logits on the raw
  # outputs of 'y', and then average across the batch.
  cross_entropy = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(y, y_))
  train_step = tf.train.GradientDescentOptimizer(0.5).minimize(cross_entropy)

  return (x,None,y,y_,train_step)


def model_MLP(nhidden):

  # Layer 1 -> 2
  x = tf.placeholder(tf.float32, [None, NDIMS])
  xnorm = x #tf.nn.batch_normalization(x, 0.25, 1., 0, 1., 0.) # shift from [0,1] to [-0.5,0.5]
  W1 = tf.Variable(tf.random_normal([NDIMS, nhidden], 0., 0.1))
  b1 = tf.Variable(tf.random_normal([nhidden], 0., 0.1))
  x2 = tf.sigmoid(tf.matmul(xnorm, W1) + b1)

  # Layer 2 -> 3
  W2 = tf.Variable(tf.random_normal([nhidden, NCLASS], 0., 0.1))
  b2 = tf.Variable(tf.random_normal([NCLASS], 0., 0.1))
  y = tf.sigmoid(tf.matmul(x2, W2) + b2)

  # Define loss and optimizer
  y_ = tf.placeholder(tf.float32, [None, 10])

  mean_squared_error = tf.reduce_mean(tf.square(y-y_))
  train_step = tf.train.GradientDescentOptimizer(0.5).minimize(mean_squared_error)

  return (x,xnorm,y,y_,train_step,W2)
  

def inspect_10worst(mnist, model, session, n_worst=10):

  (x,xnorm,y,y_,train_step) = model

  _errors = tf.nn.softmax_cross_entropy_with_logits(y, y_)
  errors = session.run(_errors, feed_dict={x: mnist.test.images,
                                           y_: mnist.test.labels})
  testfeats = mnist.test.images
  worst_feats = sorted(zip(testfeats,errors), key=lambda x: x[1], reverse=True)
  for i,(feat,error) in enumerate(worst_feats[:n_worst]):
      img = (255*feat.reshape((HEIGHT,WIDTH))).astype('uint8')
      cv2.imwrite("error-%d.png" %i, img)




def train_and_eval(mnist, model, session):

  (x,xnorm,y,y_,train_step,W2) = model

  # Train
  tf.initialize_all_variables().run()
  before = time.time()
  for _ in range(1000):
    if _%100==0:
        print _
    batch_xs, batch_ys = mnist.train.next_batch(100)
    session.run(train_step, feed_dict={x: batch_xs, y_: batch_ys})
    print (_, session.run(tf.reduce_sum(W2)))
  after = time.time()
  print 'Training took %s seconds' %(after-before)

  # Training set accuracy
  correct_prediction = tf.equal(tf.argmax(y, 1), tf.argmax(y_, 1))
  accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))
  print 'accuracy', (session.run(accuracy, feed_dict={x: mnist.train.images,
                                      y_: mnist.train.labels}))

  # Test set accuracy
  correct_prediction = tf.equal(tf.argmax(y, 1), tf.argmax(y_, 1))
  accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))
  print 'accuracy', (session.run(accuracy, feed_dict={x: mnist.test.images,
                                      y_: mnist.test.labels}))

  

def main():
  mnist = input_data.read_data_sets(FLAGS.data_dir, one_hot=True)
  #model = model_softmax()
  model = model_MLP(nhidden=300)
  session = tf.InteractiveSession()
  train_and_eval(mnist, model, session)
  inspect_10worst(mnist, model, session)



if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument('--data_dir', type=str, default='/tmp/data',
                      help='Directory for storing data')
  FLAGS = parser.parse_args()

  main()
